/*
 * CFGrammar.java
 * Copyright (C) 2004-2005, Bjorn Bringert (bringert@cs.chalmers.se)
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package se.chalmers.cs.gf.parse.chart;

import static se.chalmers.cs.gf.parse.ParseLogger.*;
import se.chalmers.cs.gf.GFException;

import java.util.*;

/**
 *  A context-free grammar.
 */
public class CFGrammar {

        private String name;

        private Category start;

        private Set<NonTerminalRule> nonTerminalRules;

        // FIXME: should maybe be a trie
        private Set<TerminalRule> terminalRules;

        private Set<EmptyRule> emptyRules;

        private Set<String> knownWords;

        public CFGrammar(String name, Category start) {
                this.name = name;
                this.start = start;
                this.nonTerminalRules = new HashSet<NonTerminalRule>();
                this.terminalRules = new HashSet<TerminalRule>();
                this.knownWords = new HashSet<String>();
                terminalRules.add(new StringLiteralRule(knownWords));
                terminalRules.add(new IntLiteralRule());
                this.emptyRules = new HashSet<EmptyRule>();
        }

        public String getName() {
                return name;
        }

        public Category getStart() {
                return start;
        }

        /**
         *  Get the rules with only non-terminals in the RHS.
         */
        public Collection<NonTerminalRule> getNonTerminalRules() {
                return nonTerminalRules;
        }

        /**
         *  Get the rules with only a single terminal in the RHS.
         */
        public Collection<TerminalRule> getTerminalRules() {
                return terminalRules;
        }

        /**
         *  Get the empty rules.
         */
        public Collection<EmptyRule> getEmptyRules() {
                return emptyRules;
        }

        /**
         *  Add a rule to the grammar.
         */
        public void addRule(Category cat, List<Symbol> rhs, String fun, int[][] profile) {
                if (rhs.isEmpty()) {
                        EmptyRule r = new EmptyRule(cat, fun, profile.length);
                        if (!emptyRules.add(r))
                                log.warning("In grammar " + name 
                                            + ": duplicate empty rule " + r);
                        log.finest("Added empty rule: " + r);
                } else {
                        // FIXME: special case for rules that are already single-terminal rule?

                        List<Category> rhsc = new LinkedList<Category>();

                        for (Symbol s : rhs) {
                                if (s instanceof Terminal) {
                                        Terminal t = (Terminal)s;
                                        Category c = addTerminal(t);
                                        rhsc.add(c);
                                } else if (s instanceof Category) {
                                        rhsc.add((Category)s);
                                } else 
                                        assert false;
                        }
                        
                        NonTerminalRule r = new NonTerminalRule(cat, rhsc, fun, profile);
                        if (!nonTerminalRules.add(r))
                                log.warning("In grammar " + name 
                                            + ": duplicate non-terminal rule " + r);
                        log.finest("Added non-terminal rule: " + r);
                }
        }

        /**
         *  Add a terminal rule.
         */
        private Category addTerminal(Terminal t) {
                Category cat = makeTerminalCategory(t);
                TerminalRule r = new TerminalRule(cat, t);
                terminalRules.add(r);
                knownWords.add(t.getToken());
                log.finest("Added terminal rule: " + r);
                return cat;
        }

        private Category makeTerminalCategory(Terminal t) {
                return new Category("$" + t.toString()); //FIXME: might have to be more advanced
        }

        public String toString() {
                StringBuilder sb = new StringBuilder();
                sb.append("grammar ").append(name).append("\n");
                sb.append("startcat ").append(start).append("\n");
                for (NonTerminalRule r : nonTerminalRules) 
                        sb.append(r.toString()).append("\n");
                for (TerminalRule r : terminalRules) 
                        sb.append(r.toString()).append("\n");
                sb.append("end grammar\n");
                return sb.toString();
        }
}
