/*
 * ChartParser.java
 * Copyright (C) 2004-2005, Bjorn Bringert (bringert@cs.chalmers.se)
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package se.chalmers.cs.gf.parse.chart;

import se.chalmers.cs.gf.abssyn.Tree;
import se.chalmers.cs.gf.parse.*;
import static se.chalmers.cs.gf.parse.ParseLogger.*;

import java.util.Collection;
import java.util.List;


/**
 *  Combines lexing, kilbury parsing and tree building.
 */
public class ChartParser implements Parser {

        private Lexer lexer;

        private CFGrammar grammar;

        private KilburyParser parser;

        public ChartParser(Lexer lexer, CFGrammar grammar) {
                this.lexer = lexer;
                this.grammar = grammar;
                parser = new KilburyParser(grammar);

                log.fine("Created chart parser for " + grammar.getName());
        }

        public Collection<Tree> parse(String str) {
                return parse(str, grammar.getStart());
        }

        public Collection<Tree> parse(String str, String cat) {
                return parse(str, new Category(cat));
        }

        public Collection<Tree> parse(String str, Category cat) {
                log.finer("Parsing \"" + str + "\" as " + cat + ".");

                List<Token> tokens = lexer.lex(str);

                Chart chart = parser.parse(tokens);

                log.finest("Finished parsing.");

                log.finest(chartInfo(chart));

                log.finest("Building trees.");

                Collection<Tree> trees = new TreeBuilder(chart).buildTree(cat);

                log.finer("Built " + trees.size() + " trees");

                return trees;
        }

        private static void ln(StringBuilder a, CharSequence s) {
                a.append(s).append('\n');
        }

        private String chartInfo(Chart chart) {
                StringBuilder s = new StringBuilder();
                ln(s, "Passive edges:");
                ln(s, chart.toString());

                ln(s, "Passive non-empty edges: ");
                int n = chart.getSize();
                for (int i = 0; i <= n; i++)
                        for (int j = i+1; j <= n; j++)
                                for (Edge e : chart.get(i,j))
                                        ln(s, "("+i+","+j+"): " + e);

                ln(s, "Complete parses:");
                for (Edge e : chart.get(0, chart.getSize()))
                        ln(s, e.toString());

                return s.toString();
        }

        public String getName() {
                return grammar.getName();
        }

        public boolean tryThisOne() {
                return true;
        }

}
