/*
 * KilburyParser.java
 * Copyright (C) 2004-2005, Bjorn Bringert (bringert@cs.chalmers.se)
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package se.chalmers.cs.gf.parse.chart;

import java.util.*;
import java.io.PrintWriter;

import se.chalmers.cs.gf.util.MultiMap;
import se.chalmers.cs.gf.parse.Token;


/**
 *  A bottom-up chart parser.
 */
public class KilburyParser {

        private CFGrammar grammar;

        /** Maps categories to the rules starting with each category. */
        private MultiMap<Category,NonTerminalRule> rulesByFirstCat;

        public KilburyParser(CFGrammar grammar) {
                this.grammar = grammar;

                this.rulesByFirstCat = new MultiMap<Category,NonTerminalRule>();
                for (NonTerminalRule r : grammar.getNonTerminalRules())
                        rulesByFirstCat.put(r.getRhs().get(0), r);
        }

        public Chart parse(List<Token> input) {
                int n = input.size();
                Chart chart = new Chart(n);

                // process first node without adding any tokens
                handleNode(chart, 0);

                int k = 0;
                for (Token t : input) {
                        k++;
                        // add terminal edges for token between k-1 and k
                        scan(chart, t, k);
                        // create all edges ending at k
                        handleNode(chart, k);
                }
                assert k == n : k + ", " + n;

                chart.purgeActive();

                return chart;
        }

        /**
         *  Use the scan rule to add edges for the token between
         *  nodes k-1 and k.
         */
        private void scan(Chart chart, Token t, int k) {
                // FIXME: inefficient to go through all terminal rules,
                //        we could use a trie to find matching rules.
                for (TerminalRule r : grammar.getTerminalRules())
                        if (r.matches(t))
                                chart.add(new TerminalEdge(r, t), k-1, k);
        }

        /**
         *  Do all processing for a node for which the terminal ending at that
         *  node has been added.
         *  Adds edges for all empty rules and iterates predict and combine 
         *  to the least fixpoint.
         */
        private void handleNode(Chart chart, int k) {
                empty(chart, k);
                while (step(chart,k)) { }
        }

        /**
         *  Add edges for all empty rules at node k.
         */ 
        private void empty(Chart chart, int k) {
                // FIXME: we could keep a set of all the EmptyRules around and 
                // link to if from the chart
                for (EmptyRule r : grammar.getEmptyRules())
                        chart.add(new EmptyEdge(r), k, k);
        }

        /**
         *  Use the predict and combine rules once on all edges ending at k.
         *  Should be iterated to a fixed-point.
         *
         * @return true if a new edge was added, false otherwise.
         */ 
        private boolean step(Chart chart, int k) {
                boolean addedNew = false;
                for (int j = 0; j <= k; j++) {
                        // FIXME: hack to avoid concurrent modification in the predict rule
                        Set<Edge> s = new HashSet<Edge>(chart.get(j,k));
                        for (Edge e : s) {
                                if (!e.isActive()) {
                                        if (predict(chart, e, j, k))
                                                addedNew = true;
                                        if (combine(chart, e, j, k))
                                                addedNew = true;
                                }
                        }
                }
                return addedNew;
        }

        /**
         *  Use the combine rule on a passive edge from j to k.
         *  @return true if a new edge was added, false otherwise.
         */ 
        private boolean combine(Chart chart, Edge e, int j, int k) {
                boolean addedNew = false;
                for (int i = 0; i <= j; i++) { 
                        // FIXME: hack to avoid concurrent modification
                        Set<Edge> s = new HashSet<Edge>(chart.get(i,j));
                        for (Edge f : s) {
                                if (f.isActive() && f.needs(e.getCategory())) {
                                        NonTerminalEdge g = new NonTerminalEdge((NonTerminalEdge)f);
                                        g.stepForward();
                                        if (chart.add(g, i, k))
                                                addedNew = true;
                                }
                        }
                }
                return addedNew;
        }

        /**
         *  Use the predict rule on a passive edge from j to k.
         *  @return true if a new edge was added, false otherwise.
         */
        private boolean predict(Chart chart, Edge e, int j, int k) {
                boolean addedNew = false;
                for (NonTerminalRule r : rulesByFirstCat.get(e.getCategory())) {
                        NonTerminalEdge f = new NonTerminalEdge(r);
                        f.stepForward();
                        if (chart.add(f, j, k))
                                addedNew = true;
                }
                return addedNew;
        }

}
