/*
 * TreeBuilder.java
 * Copyright (C) 2004-2005, Bjorn Bringert (bringert@cs.chalmers.se)
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package se.chalmers.cs.gf.parse.chart;

import se.chalmers.cs.gf.abssyn.*;
import se.chalmers.cs.gf.util.MultiMap;

import java.util.*;

/**
 *  Builds abstract syntax trees from a parsing chart.
 */
public class TreeBuilder {

        /**
         *  Keeps edges by category to speed up iteration through all edges of
         *  one category.
         *  FIXME: hacky: this could use some generalisation of Chart
         */
        private MultiMap<Category,Edge>[][] chart;

        /** The length of the input (size of the chart) */
        private int n;

        public TreeBuilder(Chart edgeChart) {
                n = edgeChart.getSize();

                chart = (MultiMap<Category,Edge>[][])new MultiMap<?,?>[n+1][n+1];
                for (int i = 0; i <= n; i++) {
                        for (int j = i; j <= n; j++) {
                                chart[i][j] = new MultiMap<Category,Edge>();
                                for (Edge e : edgeChart.get(i,j))
                                        chart[i][j].put(e.getCategory(), e);
                        }
                }
        }

        private Set<Edge> catEdgesBetween(Category c, int i, int j) {
                return chart[i][j].get(c);
        }

        /**
         *  Get all trees for the given category spanning the entire input.
         */
        public Set<Tree> buildTree(Category cat) {
                return buildTree(cat, 0, n, Collections.<Edge>emptySet());
        }

        /**
         *  Get all trees for the given category between nodes i and j.
         *  @param used The set of all edges already used between nodes i and j.
         */
        private Set<Tree> buildTree(Category cat, int i, int j, Set<Edge> used) {
//                 if (!cat.getName().startsWith("$")) {
//                         System.err.println("Building tree for category '" + cat + "' between " + i + " and " + j);
// //                        System.err.println("used: " + used);
//                 }

                assert i <= j : i + ", " + j;

                Set<Tree> ts = new HashSet<Tree>();

                for (Edge e : catEdgesBetween(cat, i, j)) {
//                                System.err.println("Considering " + e);

                        // FIXME: inelegant to use type case here
                        if (e instanceof EmptyEdge) {
                                EmptyRule r = ((EmptyEdge)e).getRule();
                                Tree[] children = new Tree[r.getArgCount()];
                                for (int x = 0; x < children.length; x++)
                                        children[x] = new MetaVariable();
                                ts.add(new Fun(r.getFun(), children));
                        } else if (e instanceof TerminalEdge) {
                                TerminalEdge te = (TerminalEdge)e;
                                TerminalRule r = te.getRule();
                                ts.add(r.makeTree(te.getToken()));
                        } else {
                                NonTerminalRule nr = ((NonTerminalEdge)e).getRule();
                                if (!used.contains(e)) {
                                        Set<Edge> newUsed = new HashSet<Edge>(used);
                                        newUsed.add(e);
                                        Set<List<Tree>> css = buildTrees(nr.getRhs(), i, j, newUsed);
                                        for (List<Tree> cs : css) {
                                                Tree[] children = keepNonTerminals(cs);
                                                Tree t = makeNode(nr, children);
                                                if (t != null) {
                                                        boolean addedNew = ts.add(t);
//                                                                 if (!addedNew) 
//                                                                         System.err.println(t + " has already been built");
                                                }
                                        }
                                } else {
//                                                System.err.println("We have seen " + e + " before");
                                }
                        }
                }
         
                return ts;
        }

        /**
         *  Keep all trees that correspond to non-terminal productions.
         */
        private Tree[] keepNonTerminals(List<Tree> ts) {
                List<Tree> r = new ArrayList<Tree>(ts.size());
                for (Tree t : ts)
                        if (treeIsNonTerminal(t))
                                r.add(t);
                return r.toArray(new Tree[r.size()]);
        }

        /**
         *  Check if a tree corresponds to a non-terminal production.
         *  FIXME: hackish
         */
        private boolean treeIsNonTerminal(Tree t) {
                return t.isLiteral() ||
                        (t instanceof Fun && !((Fun)t).getLabel().startsWith("$"));
        }

        /**
         *  Make an abstract syntax node for a rule given the trees for the children.
         *
         *  @param children Child trees from context free rule
         *  @return A syntax tree node, or null if the children could not be unified
         *          according to the rule profile.
         */
        private Tree makeNode(NonTerminalRule r, Tree[] children) {
                int[][] profile = r.getProfile();

                if (funIsCoersion(r.getFun())) {
                        assert children.length == 1;
                        assert profile.length == 1;
                        assert profile[0].length == 1;
                        assert profile[0][0] == 0;
                        return children[0];
                } else {
                        Tree[] childNodes = new Tree[profile.length];
                        for (int i = 0; i < childNodes.length; i++) {
                                childNodes[i] = new MetaVariable();
                                for (int x : profile[i]) {
                                        childNodes[i] = unifyTrees(childNodes[i], children[x]);
                                        if (childNodes[i] == null)
                                                return null;
                                }
                        }
                        
                        return new Fun(r.getFun(), childNodes);
                }
        }
        
        private boolean funIsCoersion(String fun) {
                return fun.equals("_");
        }

        public static Tree unifyTrees(Tree t1, Tree t2) {
                if (t1 == null || t2 == null)
                        throw new NullPointerException();
                if (t2 instanceof MetaVariable)
                        return t1;    // FIXME: hacky, would like multiple dispatch

                Tree t = t1.accept(new TreeUnifier(), t2);
//                 if (t == null)
//                         System.err.println("Can't unify " + t1 + " and " + t2);
//                 else
//                         System.err.println("unifyTrees(" + t1 + ", " + t2 + ") => " + t);

                return t;
        }

        // FIXME: add diagnostic messages when it fails
        private static class TreeUnifier implements TreeVisitor<Tree,Tree> {
                public Tree visit(Fun f, Tree t) {
                        if (!(t instanceof Fun))
                                return null;
                        Fun g = (Fun)t;
                        if (!f.getLabel().equals(g.getLabel()))
                                return null;
                        Tree[] cs1 = f.getChildren();
                        Tree[] cs2 = g.getChildren();
                        if (cs1.length != cs2.length)
                                return null;
                        Tree[] cs = new Tree[cs1.length];
                        for (int i = 0; i < cs.length; i++) {
                                Tree c = unifyTrees(cs1[i], cs2[i]);
                                if (c == null)
                                        return null;
                                cs[i] = c;
                        }
                        return new Fun(f.getLabel(), cs);
                }
                public Tree visit(IntLiteral l, Tree t) {
                        return l.equals(t) ? l : null;
                }
                public Tree visit(StringLiteral l, Tree t) {
                        return l.equals(t) ? l : null;
                }
                public Tree visit(MetaVariable v, Tree t) {
                        return t;
                }
        }


        /**
         *  Get all tree lists for the given symbols between i and j.
         *  Each element in the result is one of the ambigous parses. Each such element
         *  is a list of trees that together cover the input from i to j and where
         *  each tree corresponds to one of the given symbols.
         *  @param ss A non-empty list of symbols to match.
         *  @param used The set of all edges already used between nodes i and j.
         */
        private Set<List<Tree>> buildTrees(List<Category> ss, int i, int j, Set<Edge> used) {
                assert !ss.isEmpty();
                assert i <= j : i + ", " + j;

                Set<List<Tree>> r = new HashSet<List<Tree>>();

                // head is the first category that we are looking for
                Category head = ss.get(0);
                // tail is the rest of the categories
                List<Category> tail = ss.subList(1, ss.size());

                if (tail.isEmpty()) {
                        // we are only looking for the head category,
                        // so for each parse, we get a singleton sequence
                        for (Tree h : buildTree(head, i, j, used))
                                r.add(Arrays.asList(h));
                } else {
                        // go through all possible first edges
                        // k is the end index of the first edge
                        for (int k = i; k <= j; k++) {
                                // build all possible trees for the head category between i and k
                                Set<Edge> usedH = (k == j) ? used : Collections.<Edge>emptySet();
                                Set<Tree> hts = buildTree(head, i, k, usedH);
                                // build all possible tree lists for the tail categories
                                Set<Edge> usedT = (k == i) ? used : Collections.<Edge>emptySet();
                                Set<List<Tree>> rss = buildTrees(tail, k, j, usedT);
                                // make all combinations of head tree and the rest of the trees.
                                for (Tree h : hts) {
                                        for (List<Tree> rs : rss) {
                                                List<Tree> ts = new LinkedList<Tree>(rs);
                                                ts.add(0, h);
                                                boolean addedNew = r.add(ts);
//                                                         if (!addedNew) 
//                                                                 System.err.println(ts + " has already been added");
                                        }
                                }
                        }
                }

                return r;
        }
                        
}
