/*
 * TranslatorFactory.java
 * Copyright (C) 2004-2005, Bjorn Bringert (bringert@cs.chalmers.se)
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package se.chalmers.cs.gf.translate;

import se.chalmers.cs.gf.GFException;
import se.chalmers.cs.gf.gfcutil.*;
import se.chalmers.cs.gf.parse.*;
import se.chalmers.cs.gf.parse.chart.*;
import se.chalmers.cs.gf.linearize.*;
import se.chalmers.cs.gf.util.FileUtil;

import java.io.*;
import java.util.*;


/**
 *  Loads parsers, linearizers etc. and creates Translators from them.
 */
public class TranslatorFactory {

        private static Object instantiate(String className) {
                try {
                        Class<?> c = Class.forName(className);
                        return c.newInstance();
                } catch (ClassNotFoundException ex) {
                        throw new GFException("Error instantiating " 
                                              + className + ": " 
                                              + ex.getMessage(), ex);
                } catch (InstantiationException ex) {
                        throw new GFException("Error instantiating " 
                                              + className + ": " 
                                              + ex.getMessage(), ex);
                } catch (IllegalAccessException ex) {
                        throw new GFException("Error instantiating " 
                                              + className + ": " 
                                              + ex.getMessage(), ex);
                }
        }

        private static List<Object> instantiateAll(String[] classes) {
                List<Object> cs = new LinkedList<Object>();
                for (String n : classes) {
                        try {
                                cs.add(instantiate(n));
                        } catch (GFException ex) {
                                System.err.println(ex.toString());
                        }
                }
                return cs;
        }

        private static Reader readResource(String name) throws IOException {
                return FileUtil.readResource(name, "UTF-8");
        }

        /**
         *  Create a translator by reading properties from the given resource.
         */
        public static Translator createTranslator(String propResource) throws IOException {
                Properties prop = FileUtil.readPropertiesResource(propResource);
                return createTranslator(prop);
        }

        /**
         *  Create a translator from a set of properties.
         */
        public static Translator createTranslator(Properties prop) throws IOException {
                String name = FileUtil.getProp(prop, "name");
                String cfgm = FileUtil.getProp(prop, "cfgm");
                String[] xps = FileUtil.getListProp(prop, "extraparsers");
                String gfcm = FileUtil.getProp(prop, "gfcm");
                String[] xls = FileUtil.getListProp(prop, "extralinearizers");
                String lexer = prop.getProperty("lexer");

                return createTranslator(name, cfgm, gfcm, xps, xls, lexer);
        }

        /**
         *  Create a translator by reading grammars from the given resources.
         *  @param name The name of the translator.
         *  @param cfgm Resource to read the CFGM (parsing) grammar from.
         *  @param gfcm Resource to read the GFCM (linearization) grammar from.
         */
        public static Translator createTranslator(String name, String cfgm, String gfcm) 
                throws IOException {

                return createTranslator(name, cfgm, gfcm, new String[0], new String[0]);
        }

        /**
         *  Create a translator by reading grammars from the given resources.
         *  @param name The name of the translator.
         *  @param cfgm Resource to read the CFGM (parsing) grammar from.
         *  @param gfcm Resource to read the GFCM (linearization) grammar from.
         *  @param extraparsers Class names of extra parsers to load.
         *  @param extralinearizers Class names of extra linearizers to load.
         */
        public static Translator createTranslator(String name, 
                                                  String cfgm, 
                                                  String gfcm, 
                                                  String[] extraparsers, 
                                                  String[] extralinearizers)
                throws IOException {
                return createTranslator(name, cfgm, gfcm, extraparsers, extralinearizers, null);
        } 

        /**
         *  Create a translator by reading grammars from the given resources.
         *  @param name The name of the translator.
         *  @param cfgm Resource to read the CFGM (parsing) grammar from.
         *  @param gfcm Resource to read the GFCM (linearization) grammar from.
         *  @param extraparsers Class names of extra parsers to load.
         *  @param extralinearizers Class names of extra linearizers to load.
         *  @param lexer Class name of lexer, or null to use the default lexer.
         */
        public static Translator createTranslator(String name, 
                                                  String cfgm, 
                                                  String gfcm, 
                                                  String[] extraparsers, 
                                                  String[] extralinearizers,
                                                  String lexer) 
                throws IOException {

                Lexer lex;
                if (lexer == null) {
                        lex = new SimpleLexer();
                } else {
                        lex = (Lexer)instantiate(lexer);
                }


//                System.err.println("Creating parsers from " + cfgm + "...");
                Parsers ps = CFGParsers.readParsers(lex, readResource(cfgm));

                ps.addParser(new ParseAll(ps));

                for (Object o : instantiateAll(extraparsers)) {
                        Parser p = (Parser)o;
//                        System.err.println("Loaded extra parser " + p.getName());
                        ps.addParser(p);
                }

//                System.err.println("Creating linearizers from " + gfcm + "...");
                GFCMModules ms = GFCMModules.loadGFCM(readResource(gfcm));
                Linearizers ls = new GFCMLinearizers(ms);

                ls.addLinearizer(new LinearizeAll(ls));

                for (Object o : instantiateAll(extralinearizers)) {
                        Linearizer l = (Linearizer)o;
//                        System.err.println("Loaded extra linearizer " + l.getName());
                        ls.addLinearizer(l);
                }

                GFCModule absModule = ms.getMainAbstractModule();
                TypeAnnotator annotator = new TypeAnnotator(absModule);

                return new Translator(name, ps, annotator, ls);
        }

}
